#
# ReadingLists.pm - Perl module implementing a reading list object and
# associated methods.
#
# Author: j.p.knight@loughborough.ac.uk
#
# $Id: ReadingLists.pm,v 1.86 2005/09/14 10:05:46 jon Exp jon $
#
#   Copyright (C) 2001-2002 Loughborough University Library
#   Authors:  Jon Knight and Gary Brewerton
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

package ReadingLists;

use lib ("/usr/local/ReadingLists");
use strict;

#
# Location of this script
#
sub ReadingListsLibraryDir {
  return("/usr/local/ReadingLists");
}

#
# Location of backend scripts
#
sub BackendBinDir {
  return("/usr/local/ReadingLists/bin");
}

#
# Method for returning module code format
#
# Module Format explanation: the code format is made up of the following
# letters:
#
# Y = year digit
# D = department code letter
# S = stage code letter
# X = opaque letter
#
# Eg: Loughborough's ModuleCodeFormat is "YYDDSXXX" which allows module codes
# such as "01COB123"
#
sub ModuleCodeFormat {
    return("YYDDSXXX");
}

#
# Method for indicating whether we should use a toolbar for admin scripts
#
sub UseAdminToolbar {
    return 1;
}

#
# String required for getting a Cascading Style Sheet
#
sub CSS {
    return <<"EndOfCSSHeader";
  <link rel='stylesheet' type='text/css' href='/readinglists.css' title='Default'>
EndOfCSSHeader
}

#
# Method returning a URL for linking to an OPAC main page
#
sub OpacURL {
    return("http://opac.lboro.ac.uk");
}

#
# Method returning a skeleton URL for linking to an OPAC via ISBN/ISSN
#
sub OpacISBNURL {
    my $self = shift;
    my $isbn;
    if(ref($self)) {
	$isbn = shift || "";
    } else {
	$isbn = $self;
    }
    return("http://opac.lut.ac.uk/F?func=find-b&find_code=WIS&request=$isbn");
}

#
# Whether to always link to the OPAC irrespective of the held flag
#
sub AlwaysLinkToOPAC {
    return(0);
}

#
# Method returning the host part of an absolute URI for RL CGIs
#
sub HostURI {
    return("https://bookworm.lut.ac.uk");
}

#
# Method returning the Z39.50 host DNS name
#
sub Z3950Hostname {
    return("opac.lut.ac.uk");
}

#
# Method returning the Z39.50 port number
#
sub Z3950Port {
    return("9997");
}

#
# Method returning the Z39.50 database name
#
sub Z3950DBName {
    return("LBO01");
}

#
# Method returning the Z39.50 username
#
sub Z3950Username {
    return(""); # Z3950Username
}

#
# Method returning the Z39.50 password
#
sub Z3950Password {
    return(""); # Z3950Password
}

#
# Method returning the maximum number of Z39.50 matches to use per item
#
sub MaxZ3950Matches {
    return(10);
}

#
# Method to return the Z39.50 connection livespan
#
sub Z3950ConnectionLifespan {
    return(50);
}

#
# Whether to always link to the OpenURL resolver irrespective of the openurl
# flag.
#
sub AlwaysLinkToOpenURL {
    return(0);
}

#
# Method to return an OpenURL Base URL section
#
sub OpenURLBase {
    return("http://sfx.lboro.ac.uk:9003/sfx_local?");
#    return("");
}

#
# Method to return an OpenURL API Base URL section
#
sub OpenURLAPIBase {
    return("http://sfx.lboro.ac.uk:9003/SFX_API/sfx_local");
}

#
# Method to return the OpenURL link text/HTML fragment
#
sub OpenURLText {
#    return("OpenURL");
    return("<img src=\"http://sfx.lboro.ac.uk:9003/sfx_local/sfx.gif\" alt=\"SFX OpenURL link\" border=0>");
#    return ("");
}

#
# Method to return the a boolean as to whether OpenURL links should have a new
# window opened (if the browser supports that).
#
sub OpenURLNewWindow {
    return(1);
}

#
# Method to return the number of days ago that items are no longer considered
# "new"
#
sub NewItemDays {
    return(10);
}

#
# Method returning CGI path to end user CGI scripts
#
sub UserCGI {
    return("/rl");
}

#
# Method returning filesystem path to end user CGI scripts
#
sub UserCGIBinDir {
  return("/home/readinglists/cgi-bin/readinglists");
}

#
# Method returning CGI path to admin user CGI scripts
#
sub AdminCGI {
    return("/rl/admin");
}

#
# Method returning filesystem path to end admin CGI scripts
#
sub AdminCGIBinDir {
  return("/home/readinglists/cgi-bin/readinglistsadmin");
}

#
# Method returning CGI path to librarian CGI scripts
#
sub LibrarianCGI {
    return("/rl/admin/librarians");
}

#
# Method returning filesystem path to end librarian CGI scripts
#
sub LibrarianCGIBinDir {
  return("/home/readinglists/cgi-bin/readinglistslibrarian");
}

#
# Method returning CGI path to librarian CGI scripts
#
sub DefaultLibraryEmail {
    return("library\@lut.ac.uk");
}

#
# Method returning CGI path to librarian CGI scripts
#
sub SystemAdminEmail {
    return("librarysystems\@lut.ac.uk");
}

#
# Subroutine to store the database access information for readinglists
#
sub dsn {
  return "DBI:mysql:database=readinglist;user=root;password=mysqlpass;"
    . "mysql_socket=/var/lib/mysql/mysql.sock";
}

#
# Subroutine to hide HTML tag for opening debugging information
#
sub bl {
  return "<p align=center><font color=#ff0000><b>";
}

#
# Subroutine to hide HTML tag for closing debugging information
#
sub br {
  return "</b></font></p>";
}

#
# Subroutine contaning the HTML trailer for reading list CGI scripts
#
sub trailer {
  return "<p align=right>\n"
     . "</body>\n</html>\n";
}

#
# Subroutine to create a new reading list object
#
sub new {
    my $class = shift;
    my $self = {};
    bless $self, $class;
#    $self->_initialize();

    $self->{db} = DBI->connect(ReadingLists::dsn);
    return(undef) unless(defined($self->{db}));

    return $self;
}

#
# Subroutine to be called as an object destructed
#
sub DESTROY {
    my $self = shift;
    $self->{db}->disconnect if(defined($self->{db}));
}

#
# Subroutine to provide details on the available reading lists.
#
sub available_lists {
    my $self = shift;
    my @listdetails;

    my $db = $self->{db};
    return() unless(defined($db)); 

    my $sql = "select list_id, module_code, module_name, created_time, " .
      "modified_time, last_item_update, year, dept_code, stage, code, " .
      " publish from readinglist order by module_name";
    my $sh = $db->prepare($sql);
    $sh->execute;

    my(@row);
    my($count) = 0;
    while(@row = $sh->fetchrow_array) {
	my($list_id, $module_code, $module_name, $created_time,
           $modified_time, $last_time_update, $year, $dept_code,
           $stage, $code, $publish) = @row;
        $listdetails[$count]->{list_id} = $list_id;
        $listdetails[$count]->{module_code} = $module_code;
        $listdetails[$count]->{module_name} = $module_name;
        $listdetails[$count]->{created_time} = $created_time;
        $listdetails[$count]->{modified_time} = $modified_time;
        $listdetails[$count]->{last_time_update} = $last_time_update;
        $listdetails[$count]->{year} = $year;
        $listdetails[$count]->{dept_code} = $dept_code;
        $listdetails[$count]->{stage} = $stage;
        $listdetails[$count]->{code} = $code;
        $listdetails[$count]->{publish} = $publish;
        $count++;
    }
    $sh->finish;
    return(@listdetails);
}

#
# Subroutine to create a reading list
#
sub createlist {
  my $self = shift;
  my ($modulecode,$modulename,$me,$ownername,$owneremail,$ownertype,
    $owneruserid,$year,$dept_code,$stage,$code,$publish) = @_;
  my($sql,$sh,$now,$listid);

  my $db = $self->{db};
  return() unless(defined($db)); 

  if(($year eq "") || ($dept_code eq "") || ($stage eq "") || ($code eq "")) {
      return undef;
  }

  # Get the current date/time
  $modulename = $db->quote($modulename);
  $modulecode = $db->quote($modulecode);
  $ownername = $db->quote($ownername);
  $owneremail = $db->quote($owneremail);
  $ownertype = $db->quote($ownertype);
  $owneruserid = $db->quote($owneruserid);
  $year = $db->quote($year);
  $dept_code = $db->quote($dept_code);
  $stage = $db->quote($stage);
  $code = $db->quote($code);
  $publish = $db->quote($publish);

  # Create the reading list entry for this module
  $sql = "insert into readinglist (module_code, module_name, " .
        "created_time, modified_time, year, dept_code, stage, code, " .
        "publish ) values ($modulecode, $modulename, now(), now(), " .
        "$year, $dept_code, $stage, $code, $publish)";
  $sh = $db->prepare($sql);
  $sh->execute;
  $sh->finish;

  # Jiggery pokery to get the list id key back out...
  $sql = "select last_insert_id()";
  $sh = $db->prepare($sql);
  $sh->execute;
  $listid = ($sh->fetchrow_array)[0];

  # Drop an entry in the owners list for this person/list
  $sql = "insert into owner (list_id, name, userid, type, email, "
    . "modified_time) values ($listid, $ownername, $owneruserid, "
    . "$ownertype, $owneremail, now())";
  warn "Dropping ''$sql'' into owner table\n";
  $sh = $db->prepare($sql);
  $sh->execute;
  $sh->finish;

  return($listid);
}

#
# Subroutine to update a reading list
#
sub updatelist {
  my $self = shift;
  my ($modulecode,$modulename,$me,$ownername,$owneremail,$ownertype,
    $owneruserid,$year,$dept_code,$stage,$code,$publish) = @_;
  my($sql,$sh,$now,$listid);

  my $db = $self->{db};
  return() unless(defined($db)); 

  # Get the current date/time
  $modulename = $db->quote($modulename);
  $modulecode = $db->quote($modulecode);
  $ownername = $db->quote($ownername);
  $owneremail = $db->quote($owneremail);
  $ownertype = $db->quote($ownertype);
  $year = $db->quote($year);
  $dept_code = $db->quote($dept_code);
  $stage = $db->quote($stage);
  $code = $db->quote($code);
  $publish = $db->quote($publish);

  # Create the reading list entry for this module
  $sql = "update readinglist set module_name=$modulename, " .
    "created_time=now(), modified_time=now(), year=$year, " .
    "dept_code=$dept_code, stage=$stage, code=$code, publish=$publish " .
    " where module_code=$modulecode";
  $sh = $db->prepare($sql);
  $sh->execute;
  $sh->finish;

  # Jiggery pokery to get the list id key back out...
  $sql = "select list_id from readinglist where module_code=$modulecode";
  $sh = $db->prepare($sql);
  $sh->execute;
  $listid = ($sh->fetchrow_array)[0];
  $sh->finish;

  my @owners = $self->list_owners($listid);
  my $updateowner = 0;
  foreach my $owner (@owners) {
    if($owner->{userid} eq $owneruserid) {
      $updateowner = 1;
      last;
    }
  }  
  $owneruserid = $db->quote($owneruserid);

  if($updateowner) {
    $sql = "update owner set name=$ownername, type=$ownertype, "
      . "email=$owneremail, modified_time=now() "
      . "where list_id=$listid and userid=$owneruserid"
  } else {
    $sql = "insert into owner (list_id, name, userid, type, email, "
      . "modified_time) values ($listid, $ownername, $owneruserid, "
      . "$ownertype, $owneremail, now())";
  }
  warn "Update owners SQL = $sql\n";
  $sh = $db->prepare($sql);
  $sh->execute;
  $sh->finish;
}

#
# Subroutine to copy a reading list
#
use ReadingListsItem;
sub copylist {
    my $self = shift;
    my $oldmodulecode = shift;
    my $newmodulecode = shift;
    my $newmodulename = shift;
    my $me = shift;
    my $partbanner = shift;
    my $newyear = shift || "";
    my $newdeptcode = shift || "";
    my $newstage = shift || "";
    my $newopaquecode = shift || "";
    my $newpublish = shift || "";
  
    my($sql,$sh,$now,$listid);

    my $db = $self->{db};
    return(-1) unless(defined($db)); 
    my $user = $db->quote($me);
    # Get the current date/time
    chop($now = `date +'%Y-%m-%d %H:%M:%S'`);
    $now = $db->quote($now);;

    # Quote all the incoming data so that its safe to use with the database
    my $quotedoldmodulecode = $db->quote($oldmodulecode);
    my $quotednewmodulecode = $db->quote($newmodulecode);

    # Check to see if the new module already exists (and stop if it does!)
    $sql = "select module_name from readinglist where module_code = " .
	"$quotednewmodulecode";
    $sh = $db->prepare($sql);
    $sh->execute;
    return(-2) if($sh->rows > 0);
    $sh->finish;

    # Check that the old module exists (and exit if it doesn't)
    $sql = "select module_name from readinglist where module_code = " .
	"$quotedoldmodulecode";
    $sh = $db->prepare($sql);
    $sh->execute;
    return(-3) if($sh->rows <= 0);
    $sh->finish;

    # Jiggery pokery to get the list id key back out...
    $sql = "select list_id,publish from readinglist where module_code = " .
	"$quotedoldmodulecode";
    $sh = $db->prepare($sql);
    $sh->execute;
    my @row = ($sh->fetchrow_array);
    $listid = $row[0];
    my $publishflag = $row[1];
    $sh->finish;

    # Preserve publish flag unless user says otherwise.
    if($newpublish eq "") {
	$newpublish = $publishflag;
    }

    # Create the new reading list with one of the owners of the old list
    # and then update it with the other owners.
    my @owners = $self->list_owners($listid);
    my $first = 1;
    my $newlistid = -1;
    foreach my $owner (@owners) {
	my $ownername = $owner->{name};
	my $ownertype = $owner->{type};
	my $owneremail = $owner->{email};
	my $owneruserid = $owner->{userid};
	if($first) {
	    $newlistid = $self->createlist($newmodulecode, $newmodulename, 
                                           $me, $ownername, $owneremail,
					   $ownertype, $owneruserid,
					   $newyear,$newdeptcode,$newstage,
					   $newopaquecode,$newpublish);
	    $first = 0;
	} else {
	    $self->updatelist($newmodulecode, $newmodulename, $me, 
			      $ownername, $owneremail, $ownertype, 
			      $owneruserid,
			      $newyear,$newdeptcode,$newstage,
			      $newopaquecode,$newpublish);
	}
    }

    if($partbanner ne "") {
	print STDOUT "<p>Copied list headers - starting on items...";
    }
    # Grab the entries from the old list and then add them to the new list
    my @entries = $self->getlistentries($listid);
    my $processed = 0;
    foreach my $entry (@entries) {
	my $material_id = $entry->{material_id};
	my $note_id = $entry->{note_id};
	my $link_id = $entry->{link_id};
	my $rank = $entry->{rank};
	my $new_material_id = "";
	my $new_note_id = "";
	my $new_link_id = "";
#      warn "\nlistid = $listid\n";
#      warn "material_id = $material_id\n";
#      warn "note_id = $note_id\n";
	if($material_id) {
	    $sql = "select * from material where material_id = $material_id";
	    $sh = $db->prepare($sql);
	    $sh->execute;
	    my $r = $sh->fetchrow_hashref;;
	    my %row1 = %{$r};
	    $sh->finish;
	    $sql = "insert into material (";
	    my $comma = "";
	    foreach my $column (keys %row1) {
		next if($column eq "material_id");
		next if($column eq "");
		$sql .= "$comma$column";
		$comma = ",";
	    }
	    $sql .= ") values (";
	    $comma = "";
	    foreach my $column (keys %row1) {
		next if($column eq "material_id");
		next if($column eq "");
		if($column eq "created_time" || $column eq "modified_time") {
		    $row1{$column} = $now;
		}
		$sql .= $comma.$db->quote($row1{$column});
		$comma = ",";
	    }
	    $sql .= ")";
#          warn "SQL = $sql\n";
	    $sh = $db->prepare($sql);
	    $sh->execute;
	    $sh->finish;
	    $sql = "select LAST_INSERT_ID() from material";
	    $sh = $db->prepare($sql);
	    $sh->execute;
	    $new_material_id = $sh->fetch->[0];
	    $sh->finish;
	    if($note_id ne "") {
		my $notetext = $self->{db}->quote($self->getnote($note_id));
		$sql = "insert into annotation (note_id, content, " .
		    "created_time, created_by, modified_time, " .
		    "modified_by) values (0, $notetext, $now, $user, " .
                    "$now, $user)";
		$sh = $self->{db}->prepare($sql);
		$sh->execute;
		$sh->finish;
		$sql = "select LAST_INSERT_ID() from annotation";
		$sh = $self->{db}->prepare($sql);
		$sh->execute;
		$new_note_id = $sh->fetch->[0];
		$sh->finish;
	    }
	    $sql = "insert into sequence (list_id, material_id, " .
                 "note_id, rank, created_time, created_by, modified_time, " .
                 "modified_by, number_of_edits) values ($newlistid, " .
                 "$new_material_id, $new_note_id, $rank, $now, $user, " .
                 "$now, $user, 0)";
	    $sh = $db->prepare($sql);
	    warn "SQL = $sql\n";
	    $sh->execute;
	    $sh->finish;
	} elsif($note_id) {
	    my $notetext = $self->getnote($note_id);
	    $self->insertnote($newmodulecode, $notetext);
	} elsif($link_id) {
	    my($link_id,$modulecode,$target_module_code,$prelinktext,
	       $postlinktext,$rollover) = $self->getlink($link_id);
	    $self->insertlink($newmodulecode, $target_module_code,
			      $prelinktext, $postlinktext, $rollover);
	}
	$processed++;
	if((($processed % 10) == 0) && ($partbanner ne "")) {
	    print STDOUT "$processed... ";
	}
    }
    if($partbanner ne "") {
	print STDOUT "<b>completed $processed</b> items.</p>\n";
    }
    # Update the last_item_update field in the reading list table.  This is
    # used by some of our code to quickly scan for readinglists that have
    # been altered after a given datetime
    $sql = "update readinglist set last_item_update=$now ".
	"where module_code = '$newmodulecode'";
    $sh = $db->prepare($sql);
    $sh->execute;
    $sh->finish;
    return(0);
}

#
# Subroutine to append the contents of an existing material item to another
# reading list
#
sub append_material {
    my $self = shift;
    my $newlistid = shift;
    my $material_id = shift;
    my $me = shift;

    my $new_material_id;
    my $new_note_id;
    
    my $user = $self->{db}->quote($me);
    # Get the current date/time
    my $now;
    chop($now = `date +'%Y-%m-%d %H:%M:%S'`);
    $now = $self->{db}->quote($now);;
    my $sql = "select * from material where material_id = $material_id";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my $r = $sh->fetchrow_hashref;
    my %row1 = %{$r};
    $sh->finish;
    $sql = "insert into material (";
    my $comma = "";
    foreach my $column (keys %row1) {
	next if($column eq "material_id");
	next if($column eq "");
	$sql .= "$comma$column";
	$comma = ",";
    }
    $sql .= ") values (";
    $comma = "";
    my $note_id = "";
    foreach my $column (keys %row1) {
	next if($column eq "material_id");
	next if($column eq "");
	if($column eq "created_time" || $column eq "modified_time") {
	    $row1{$column} = $now;
	}
	if($column eq "created_by" || $column eq "modified_by") {
	    $row1{$column} = $user;
	}
	$sql .= $comma.$self->{db}->quote($row1{$column});
	$comma = ",";
    }
    $sql .= ")";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;
    $sql = "select LAST_INSERT_ID() from material";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $new_material_id = $sh->fetch->[0];
    $sh->finish;
    $sql = "select note_id from sequence where material_id=$material_id";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows > 0) {
	$note_id = $sh->fetch->[0];
    }
    if($note_id ne "") {
	my $notetext = $self->{db}->quote($self->getnote($note_id));
	$sql = "insert into annotation (note_id, content, " .
	    "created_time, created_by, modified_time, " .
	    "modified_by) values (0, $notetext, $now, $user, " .
            "$now, $user)";
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	$sh->finish;;
	$sql = "select LAST_INSERT_ID() from annotation";
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	$new_note_id = $sh->fetch->[0];
	$sh->finish;
    } else {
	$sql = "insert into annotation (note_id, content, " .
	    "created_time, created_by, modified_time, " .
	    "modified_by) values (0, '', $now, $user, " .
            "$now, $user)";
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	$sh->finish;;
	$sql = "select LAST_INSERT_ID() from annotation";
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	$new_note_id = $sh->fetch->[0];
	$sh->finish;
    }

    # The maximum rank in the destination table.  We need this because
    # we're going to stick the material at the end of the sequence.
    $sql = "select max(rank) from sequence where " .
      "list_id = $newlistid";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my $rank = 1;
    if($sh) {
        my @row1 = $sh->fetchrow_array;
        $rank = $row1[0];
        $rank++;
    }
    $sh->finish;

    $sql = "insert into sequence (list_id, material_id, " .
	"note_id, rank, created_time, created_by, modified_time, " .
	"modified_by, number_of_edits) values ($newlistid, " .
        "$new_material_id, $new_note_id, $rank, $now, $user, " .
        "$now, $user, 0)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;;

}

#
# Subroutine to append the contents of an existing note to another
# reading list
#
sub append_note {
    my $self = shift;
    my $newlistid = shift;
    my $note_id = shift;
    my $me = shift;

    # Get the user doing this insert and the current time
    my $now;
    chop($now = `date +'%Y-%m-%d %H:%M:%S'`);
    $now = $self->{db}->quote($now);
    my $user = $ENV{REMOTE_USER};
    unless (defined($user)) { chop($user = `whoami`); }
    $user = $self->{db}->quote($user);

    # Get the existing note text from the old list
    $note_id = $self->{db}->quote($note_id);
    my $sql = "select content from annotation where note_id = $note_id";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my $notetext = $sh->fetch->[0];
    $sh->finish;

    # Insert the new note into the note table.
    $notetext = $self->{db}->quote($notetext);
    $sql = "insert into annotation (content, created_time, " .
    "created_by, modified_time, modified_by) values ($notetext, " .
	"$now, $user, $now, $user)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;

    # Get the note_id of the newly created annotation
    $sql = "select LAST_INSERT_ID() from annotation";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my $new_note_id = $sh->fetch->[0];
    $sh->finish;

    # Work out what sequence ID relates to this reading list and the maximum
    # rank in the table.  We need the latter because we're going to stick 
    # the new note at the end of the sequence.
    my $rank = 0;
    $sql = "select max(rank) from sequence where " .
      "list_id = $newlistid";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh) {
        my @row1 = $sh->fetchrow_array;
        $rank = $row1[0];
        $rank++;
    }
    $sh->finish;

    # Insert the information about this item in the sequence table.  We
    # need to get a ranking for this item - it'll be one higher than the
    # previous highest value to ensure that it is at the end of the list
    $sql = "insert into sequence (list_id, " .
      "note_id, rank, created_time, created_by, modified_time, " .
      "modified_by, number_of_edits) values ($newlistid, " .
      "$new_note_id, $rank, $now, $user, $now, $user, 0)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;

    return($new_note_id);
}

#
# Subroutine to append the contents of an existing link to another
# reading list
#
sub append_link {
    my $self = shift;
    my $newlistid = shift;
    my $link_id = shift;
    my $me = shift;

    # Get the user doing this insert and the current time
    my $now;
    chop($now = `date +'%Y-%m-%d %H:%M:%S'`);
    $now = $self->{db}->quote($now);
    my $user = $ENV{REMOTE_USER};
    unless (defined($user)) { chop($user = `whoami`); }
    $user = $self->{db}->quote($user);

    # Get the existing note text from the old list
    $link_id = $self->{db}->quote($link_id);
    my $sql = "select target_list_id, rollover, prelinktext, postlinktext " .
	" from link where link_id = $link_id";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my @row = $sh->fetchrow_array;
    my $target_listid = $row[0];
    my $rollover = $row[1];
    my $prelinktext = $row[2];
    my $postlinktext = $row[3];
    $sh->finish;

    # Insert the new link into the note table.
    $rollover = $self->{db}->quote($rollover);
    $prelinktext = $self->{db}->quote($prelinktext);
    $postlinktext = $self->{db}->quote($postlinktext);
    $sql = "insert into link (link_id, source_list_id, target_list_id, ".  
	"rollover, prelinktext, postlinktext, created_time, " .
	"created_by, modified_time, modified_by) values (0, $newlistid, " .
        "$target_listid, $rollover, $prelinktext, $postlinktext, " .
	"$now, $user, $now, $user)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;

    # Get the link_id of the newly created annotation
    $sql = "select LAST_INSERT_ID() from link";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my $new_link_id = $sh->fetch->[0];
    $sh->finish;

    # Get the maximum  rank in the sequence table.  We need this because
    # we're going to stick the new link at the end of the sequence.
    my $rank = 0;
    $sql = "select max(rank) from sequence where " .
      "list_id = $newlistid";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh) {
        my @row1 = $sh->fetchrow_array;
        $rank = $row1[0];
        $rank++;
    }
    $sh->finish;

    # Insert the information about this item in the sequence table.  We
    # need to get a ranking for this item - it'll be one higher than the
    # previous highest value to ensure that it is at the end of the list
    $sql = "insert into sequence (list_id, " .
      "link_id, rank, created_time, created_by, modified_time, " .
      "modified_by, number_of_edits) values ($newlistid, " .
      "$new_link_id, $rank, $now, $user, $now, $user, 0)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;

    return($new_link_id);
}

#
# Subroutine to update link in a set of reading lists (at year end rollover)
#
sub rolloverlinks {
    my $self = shift;
    my $newmodulecode = shift;
    my $newyear = shift || "";
    my $newdeptcode = shift || "";
    my $newstage = shift || "";
    my $newopaquecode = shift || "";
    my $newpublish = shift || "";

    my $qnewmodulecode = $self->{db}->quote($newmodulecode);
    my $sql = "select link.link_id from link, readinglist, sequence " .
	"where readinglist.module_code = $qnewmodulecode " .
	"and sequence.list_id = readinglist.list_id " .
	"and link.link_id = sequence.link_id " .
	"and link.rollover = 'Y'";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;

    # Loop round and update the target_list_id in the link table by extracting
    # the existing one, working out what old module code that pointed to,
    # updating this module code with the year part of the module code that
    # we're looking at and then reversing the process to find a new
    # target_list_id from the newly generated target module code and updating
    # the link table appropriately
    while(my $res = $sh->fetchrow_arrayref) {
        my $link_id = $res->[0];
        my $nsql = "select target_list_id from link where link_id = $link_id";
        my $nsh = $self->{db}->prepare($nsql);
        $nsh->execute;
        my $target_list_id = $nsh->fetch->[0];
        $nsh->finish;
        $nsql = "select module_code from readinglist where " .
	    "list_id = $target_list_id";
        $nsh = $self->{db}->prepare($nsql);
        $nsh->execute;
        my $target_module_code = $nsh->fetch->[0];
        $nsh->finish;
        ######################################################################
        # This needs fixing to use years and not rely on Lboro module format #
        ######################################################################
	substr($target_module_code,0,2,substr($newmodulecode, 0, 2));
        $target_module_code = $self->{db}->quote($target_module_code);
        $nsql = "select list_id from readinglist where " .
	    "module_code = $target_module_code";
        $nsh = $self->{db}->prepare($nsql);
        $nsh->execute;
        $target_list_id = $nsh->fetch->[0];
        $nsh->finish;
        $nsql = "update link set target_list_id = $target_list_id " .
	    "where link_id = $link_id";
        $nsh = $self->{db}->prepare($nsql);
        $nsh->execute;
        $nsh->finish;
    }
    $sh->finish;
}

#
# Subroutine to return details of departments with reading lists
#
sub departments {
    my $self = shift;
    my $sortfield = shift || "";
    my @results = ();

    my $orderby = " order by dept_name";
    if($sortfield eq "faculty") {
	$orderby = " order by dept_faculty";
    } elsif($sortfield eq "facultyname") {
	$orderby = " order by dept_faculty, dept_name";
    } elsif($sortfield eq "code") {
	$orderby = " order by dept_code";
    }
    my $db = $self->{db};
    return() unless(defined($db)); 

    my $sql = "select dept_id, dept_code, dept_name, dept_faculty " .
	"from departments where browsable = 'Y' $orderby";
    my $sh = $db->prepare($sql);
    $sh->execute;
    return(undef) if($sh->rows < 1);
    my @row;
    my $count = 0;
    while(@row = $sh->fetchrow_array) {
         $results[$count]->{dept_id} = $row[0];
         $results[$count]->{dept_code} = $row[1];
         $results[$count]->{dept_name} = $row[2];
         $results[$count]->{dept_faculty} = $row[3];
         $count++;
    }
    return(@results);
}

#
# Subroutine to check the validity of a module code
#
sub validmodulecode {
    my $self = shift;
    my($modulecode) = @_;

    my $modulecodeformat = ReadingLists::ModuleCodeFormat;

    # Check that the opaque positions are OK
    my $opaqueregexp = $modulecodeformat;
    $opaqueregexp =~ s/[^X]/\./g;
    $opaqueregexp =~ s/X/\[a-zA-Z0-9]/g;
    if($modulecode !~ /^$opaqueregexp$/) {
        warn "Failed Opaque Test\n";
	return(0);
    }

    # Check that the year positions are numeric
    my $yearregexp = $modulecodeformat;
    $yearregexp =~ s/[^Y]/\./g;
    $yearregexp =~ s/Y/\\d/g;
    if($modulecode !~ /^$yearregexp$/) {
        warn "Failed Year Test\n";
	return(0);
    }

    # Check that the department positions are in the module code and that
    # the department code exists (assuming we have 'D's in the module code
    # format of course!)
    my $deptregexp = $modulecodeformat;
    $deptregexp =~ s/[^D]/\./g;
    $deptregexp =~ s/D/\[a-zA-Z0-9]/g;
    $deptregexp =~ s/^([^\[]*)\[/$1\(\[/;
    $deptregexp =~ s/\]([^\]]*)$/\]\)$1/;
    if($modulecode !~ /^$deptregexp$/) {
        warn "Failed Dept Test 1\n";
	return(0);
    }
    my $deptcode = $1;
    if($modulecodeformat =~ /D/) {
	my $db = $self->{db};
	$deptcode = $db->quote($deptcode);
	my $sql = "select dept_code from departments where " .
	    "strcmp(dept_code, $deptcode)=0";
	my $sth = $db->prepare($sql);
	$sth->execute;
        if($sth->rows < 1) {
	    warn "Failed Dept Test 2\n";
	    $sth->finish;
            return(0);
	}
	$sth->finish;
    }

    # Check stage positions are in the module code and check the resulting
    # stage code against the database if appropriate.
    my $stageregexp = $modulecodeformat;
    $stageregexp =~ s/[^S]/\./g;
    $stageregexp =~ s/S/\[a-zA-Z0-9]/g;
    $stageregexp =~ s/^([^\[]*)\[/$1\(\[/;
    $stageregexp =~ s/\]([^\]]*)$/\]\)$1/;
    if($modulecode !~ /^$stageregexp$/) {
        warn "Failed Stage Test 1\n";
	return(0);
    }
    my $stagecode = $1;
    if($modulecodeformat =~ /S/) {
	my $db = $self->{db};
	$stagecode = $db->quote($stagecode);
	my $sql = "select code from stage where " .
	    "strcmp(code, $stagecode)=0";
	my $sh = $db->prepare($sql);
	$sh->execute;
        if($sh->rows < 1) {
            warn "Failed Stage Test 1\n";
            return(0);
        }
	$sh->finish;
    }

    # Passed all the tests: Jobs A Good Un!
    return(1);
}

#
# Subroutine to provide details of owners of a reading list
#
sub list_owners {
    my $self = shift;
    my $list_id = shift;
    my @owners;

    return() if(!defined($list_id) || $list_id eq "");
    my $sql = "select name, userid, type, email from owner where " .
	  "list_id = $list_id";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    return() if($sh->rows < 1);

    my (@row,$count);
    $count = 0;
    while(@row = $sh->fetchrow_array) {
	$owners[$count]->{name} = $row[0];
	$owners[$count]->{userid} = $row[1];
	$owners[$count]->{type} = $row[2];
	$owners[$count]->{email} = $row[3];
	$count++;
    }
    $sh->finish;
    return(@owners);
}

#
# Subroutine to lookup details of items in all reading lists with a
# particular control number
#
sub controlfind {
    my $self = shift;
    my($control) = @_;
    my @materials;

    $control = $self->{db}->quote($control);
    my $sql = "select material_id from material ".
      "where control_no = $control";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    return undef if($sh->rows < 1);
    my $count = 0;
    my $material_id;
    while(($material_id)= $sh->fetchrow_array) {
        $materials[$count]->{material_id} = $material_id;
        $sql = "select list_id from sequence where material_id " .
           "= $material_id";
        my $nextsh = $self->{db}->prepare($sql);
        $nextsh->execute;
        my($list_id) = $nextsh->fetchrow_array;
        $materials[$count]->{list_id} = $list_id;
        $nextsh->finish;
        $sql = "select module_code, module_name from readinglist " .
          "where list_id = $list_id";
        $nextsh = $self->{db}->prepare($sql);
        $nextsh->execute;
        my(@row) = $nextsh->fetchrow_array;
        $materials[$count]->{module_code} = $row[0];
        $materials[$count]->{module_name} = $row[1];
        $nextsh->finish;
        $count++;
     }
     return(@materials);
}

#
# Subroutine to give details about a specific reading list given a module
# code.
#
sub list_details {
    my $self = shift;
    my $modulecode = $self->{db}->quote(shift);
    my $result;

    my $sql = "select list_id, module_name, created_time, modified_time, ".
      " last_item_update, year, dept_code, stage, code, " .
      " publish from readinglist where module_code = $modulecode";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    return undef if($sh->rows < 1);
    my @row = $sh->fetchrow_array;
    $result->{list_id} = $row[0];
    $result->{module_name} = $row[1];
    $result->{created_time}  = $row[2];
    $result->{modified_time} = $row[3];
    $result->{last_item_update} = $row[4];
    $result->{year} = $row[5];
    $result->{dept_code} = $row[6];
    $result->{stage} = $row[7];
    $result->{code} = $row[8];
    $result->{publish} = $row[9];
    $sh->finish;
    return($result);
}

#
# Subroutine to return the ids of the materials/notes in the reading list
#
sub getlistentries {
    my $self = shift;
    my $list_id = shift;
    my $sortfield = shift || "";
    my @results;

    my $sql;

    if($sortfield eq "authors") { 
	$sql = "select sequence.sequence_id, sequence.material_id, " .
               "sequence.note_id, sequence.link_id, sequence.rank, " .
               "sequence.created_time, " .
               "sequence.created_by, sequence.modified_time, " .
               "sequence.modified_by " .
	       "from sequence, material " .
               "where sequence.list_id = $list_id " .
               "and material.material_id = sequence.material_id " .
               "order by concat(material.partauthor, material.authors, " .
	       "material.parttitle, material.title)";
    } elsif($sortfield eq "pubdate") { 
	$sql = "select sequence.sequence_id, sequence.material_id, " .
               "sequence.note_id, sequence.link_id, sequence.rank, " .
               "sequence.created_time, " .
               "sequence.created_by, sequence.modified_time, " .
               "sequence.modified_by " .
	       "from sequence, material " .
               "where sequence.list_id = $list_id " .
               "and material.material_id = sequence.material_id " .
               "order by material.pubdate desc, " .
               "concat(material.partauthor, material.authors, " .
	       "material.parttitle, material.title)";
    } else {
	$sql = "select sequence_id, material_id, note_id, link_id, rank, " .
               "created_time, created_by, modified_time, modified_by " .
	       "from sequence where list_id = $list_id order by rank";
    }
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    @results = ();
    if ($sh->rows >= 1) {
        my @row;
        my $count = 0;
        while(@row = $sh->fetchrow_array) {
            $results[$count]->{sequence_id} = $row[0];
            $results[$count]->{material_id} = $row[1];
            $results[$count]->{note_id} = $row[2];
            $results[$count]->{link_id} = $row[3];
            $results[$count]->{rank} = $row[4];
            $results[$count]->{created_time} = $row[5];
            $results[$count]->{create_by} = $row[6];
            $results[$count]->{modified_time} = $row[7];
            $results[$count]->{modified_by} = $row[8];
            $count++;
        }
        $sh->finish;
        return(@results);
    }
    $sh->finish;
    return(@results);
}

#
# Subroutine to get a note from a reading list given its note_id;
#
sub getnote {
    my $self = shift;
    my $note_id = shift;
    my $notetext = "";

    if(defined($note_id)) {
	my $sql = "select content from annotation where note_id = $note_id";
	my $sh = $self->{db}->prepare($sql);
	$sh->execute;
	$notetext = $sh->fetch->[0] || "";
        $sh->finish;
    }
    return($notetext);
}

#
# Subroutine to insert a new standalone annotation into the reading list
#
sub insertnote {
    my $self = shift;
    my ($modulecode, $notetext) = @_;
    my $note_id = undef;

    $modulecode = $self->{db}->quote($modulecode);
    my $sql = "select list_id from readinglist ".
      "where module_code = $modulecode";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows < 1){
	$sh->finish;
	return(undef) ;
    }
    my ($listid) = $sh->fetch->[0];
    $sh->finish;

    # Get the user doing this insert and the current time
    my $now;
    chop($now = `date +'%Y-%m-%d %H:%M:%S'`);
    $now = $self->{db}->quote($now);
    my $user = $ENV{REMOTE_USER};
    unless (defined($user)) { chop($user = `whoami`); }
    $user = $self->{db}->quote($user);

    # Insert the new note into the note table.
    $notetext = $self->{db}->quote($notetext);
    $sql = "insert into annotation (note_id, content, created_time, " .
    "created_by, modified_time, modified_by) values (0, $notetext, " .
	"$now, $user, $now, $user)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;

    # Get the note_id of the newly created annotation
    $sql = "select LAST_INSERT_ID() from annotation";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $note_id = $sh->fetch->[0];
    $sh->finish;

    # Work out what sequence ID relates to this reading list and the maximum
    # rank in the table.  We need the latter because we're going to stick 
    # the new note at the end of the sequence.
    $sql = "select max(sequence_id) from sequence where " .
      "list_id = $listid";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my $seqid = 0;
    my $rank = 0;
    if($sh) {
        my @row1 = $sh->fetchrow_array;
        $seqid = $row1[0];
        $seqid++;
    }
    $sh->finish;
    $sql = "select max(rank) from sequence where " .
      "list_id = $listid";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh) {
        my @row1 = $sh->fetchrow_array;
        $rank = $row1[0];
        $rank++;
    }
    $sh->finish;

    # Insert the information about this item in the sequence table.  We
    # need to get a ranking for this item - it'll be one higher than the
    # previous highest value to ensure that it is at the end of the list
    $sql = "insert into sequence (list_id, " .
      "note_id, rank, created_time, created_by, modified_time, " .
      "modified_by, number_of_edits) values ($listid, " .
      "$note_id, $rank, $now, $user, $now, $user, 0)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;

    return($note_id);
}

#
# Subroutine to update an existing standalone annotation in the reading list
#
sub updatenote {
    my $self = shift;
    my ($note_id, $notetext) = @_;

    # Get the user doing this insert and the current time
    my $now;
    chop($now = `date +'%Y-%m-%d %H:%M:%S'`);
    $now = $self->{db}->quote($now);
    my $user = $ENV{REMOTE_USER};
    unless (defined($user)) { chop($user = `whoami`); }
    $user = $self->{db}->quote($user);

    $notetext = $self->{db}->quote($notetext);
    $note_id = $self->{db}->quote($note_id);

    # Updete content and the modification time/user in the annotation
    my $sql = "update annotation set content = $notetext, " . 
      "modified_by = $user, modified_time = $now " .
      "where note_id = $note_id";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;

    # Updete the modification time/user in the sequence
    $sql = "update sequence set modified_by = $user, " .
      "modified_time = $now where note_id = $note_id";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;

}

#
# Subroutine to get a link from a reading list given its link_id;
#
sub getlink {
    my $self = shift;
    my $link_id = shift;
    my $source_module_code = "";
    my $target_module_code = "";
    my $prelinktext = "";
    my $postlinktext = "";
    my $rollover = "";

    if(defined($link_id)) {
        $link_id = $self->{db}->quote($link_id);
	my $sql = "select source_list_id, target_list_id, prelinktext, " .
	    "postlinktext, rollover from link where link_id = $link_id";
	my $sh = $self->{db}->prepare($sql);
	$sh->execute;
	my $linkdata = $sh->fetchrow_arrayref || ();
        my $source_list_id = $linkdata->[0];
        my $target_list_id = $linkdata->[1];
        $prelinktext = $linkdata->[2];
        $postlinktext = $linkdata->[3];
        $rollover = $linkdata->[4];
        $sh->finish;
        $sql = "select module_code from readinglist where list_id = ".
	    $source_list_id;
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	$source_module_code = $sh->fetch->[0] || "";
        $sh->finish;
        $sql = "select module_code from readinglist where list_id = ".
	    $target_list_id;
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	if($sh->rows) {
	    $target_module_code = $sh->fetch->[0];
	} else {
	    $target_module_code = "";
	}
        $sh->finish;
    }

    return($link_id, $source_module_code, $target_module_code,
           $prelinktext, $postlinktext, $rollover);
}

#
# Subroutine to insert a new link into the reading list
#
sub insertlink {
    my $self = shift;
    my ($modulecode, $target_module_code, $prelinktext, $postlinktext,
	$rollover) = @_;
    my $link_id = undef;

    # Sanitize the parameters we've been given.
    $modulecode = $self->{db}->quote($modulecode);
    $target_module_code = $self->{db}->quote($target_module_code);
    $rollover = $self->{db}->quote($rollover);
    $prelinktext = $self->{db}->quote($prelinktext);
    $postlinktext = $self->{db}->quote($postlinktext);

    # Get the list ID for the source reading list.
    my $sql = "select list_id from readinglist ".
      "where module_code = $modulecode";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows < 1){
	$sh->finish;
	return(undef) ;
    }
    my ($listid) = $sh->fetch->[0];
    $sh->finish;

    # Get the list ID for the source reading list.
    $sql = "select list_id from readinglist ".
      "where module_code = $target_module_code";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows < 1){
	$sh->finish;
	return(undef) ;
    }
    my ($target_listid) = $sh->fetch->[0];
    $sh->finish;

    # Get the user doing this insert and the current time
    my $now;
    chop($now = `date +'%Y-%m-%d %H:%M:%S'`);
    $now = $self->{db}->quote($now);
    my $user = $ENV{REMOTE_USER};
    unless (defined($user)) { chop($user = `whoami`); }
    $user = $self->{db}->quote($user);

    # Insert the new link into the note table.
    $sql = "insert into link (link_id, source_list_id, target_list_id, ".  
	"rollover, prelinktext, postlinktext, created_time, " .
	"created_by, modified_time, modified_by) values (0, $listid, " .
        "$target_listid, $rollover, $prelinktext, $postlinktext, " .
	"$now, $user, $now, $user)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;

    # Get the link_id of the newly created link
    $sql = "select LAST_INSERT_ID() from link";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $link_id = $sh->fetch->[0];
    $sh->finish;

    # Work out what sequence ID relates to this reading list and the maximum
    # rank in the table.  We need the latter because we're going to stick 
    # the new link at the end of the sequence.
    $sql = "select max(sequence_id) from sequence where " .
      "list_id = $listid";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my $seqid = 0;
    my $rank = 0;
    if($sh) {
        my @row1 = $sh->fetchrow_array;
        $seqid = $row1[0];
        $seqid++;
    }
    $sh->finish;
    $sql = "select max(rank) from sequence where " .
      "list_id = $listid";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh) {
        my @row1 = $sh->fetchrow_array;
        $rank = $row1[0];
        $rank++;
    }
    $sh->finish;

    # Insert the information about this link in the sequence table.  We
    # need to get a ranking for this item - it'll be one higher than the
    # previous highest value to ensure that it is at the end of the list
    $sql = "insert into sequence (list_id, " .
      "link_id, rank, created_time, created_by, modified_time, " .
      "modified_by, number_of_edits) values ($listid, " .
      "$link_id, $rank, $now, $user, $now, $user, 0)";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;

    return($link_id);
}

#
# Subroutine to update an existing link in the reading list
#
sub updatelink {
    my $self = shift;
    my ($link_id,$prelinktext,$postlinktext,$rollover) = @_;

    # Get the user doing this insert and the current time
    my $now;
    chop($now = `date +'%Y-%m-%d %H:%M:%S'`);
    $now = $self->{db}->quote($now);
    my $user = $ENV{REMOTE_USER};
    unless (defined($user)) { chop($user = `whoami`); }
    $user = $self->{db}->quote($user);

    $link_id = $self->{db}->quote($link_id);
    $prelinktext = $self->{db}->quote($prelinktext);
    $postlinktext = $self->{db}->quote($postlinktext);
    $rollover = $self->{db}->quote($rollover);

    # Updete content and the modification time/user in the annotation
    my $sql = "update link set prelinktext = $prelinktext, " . 
      "postlinktext = $postlinktext, rollover = $rollover, " .
      "modified_by = $user, modified_time = $now " .
      "where link_id = $link_id";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;

    # Updete the modification time/user in the sequence
    $sql = "update sequence set modified_by = $user, " .
      "modified_time = $now where link_id = $link_id";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;
}

#
# Subroutine to get the details of reading lists owned by a particular
# user.
#
sub listsowned {
  my $self = shift;
  my $user = shift;
  my @result;

  @result = undef;
  $user = $self->{db}->quote($user);
  my $sql = "select readinglist.module_code, readinglist.module_name, "
   . "readinglist.created_time, readinglist.modified_time, "
   . "readinglist.list_id, readinglist.year, readinglist.dept_code, " 
   . "readinglist.stage, readinglist.code, readinglist.publish, "
   . "readinglist.last_item_update, readinglist.last_item_update_by "
   . " from readinglist, owner where owner.userid = $user "
   . " and owner.list_id = readinglist.list_id "
   . "order by readinglist.module_code";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  my $count = 0;
  my @row1;
  while(@row1 = $sh->fetchrow_array) {
      $result[$count]->{module_code} = $row1[0];
      $result[$count]->{module_name} = $row1[1];
      $result[$count]->{created_time} = $row1[2];
      $result[$count]->{modified_time} = $row1[3];
      $result[$count]->{list_id} = $row1[4];
      $result[$count]->{year} = $row1[5];
      $result[$count]->{dept_code} = $row1[6];
      $result[$count]->{stage} = $row1[7];
      $result[$count]->{code} = $row1[8];
      $result[$count]->{publish} = $row1[9];
      $result[$count]->{last_item_update} = $row1[10];
      $result[$count]->{last_item_update_by} = $row1[11];
      $count++;
  }
  $sh->finish;
  return(@result);
} 

#
# Subroutine to reorder a material item in a reading list
#
sub reorderitem {
  my $self = shift;
  my($material_id,$newrank) = @_;
  my $sequence_id;

  $material_id = $self->{db}->quote($material_id);
  my $sql = "select sequence_id from sequence where " .
    "material_id = $material_id";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows < 1) {
      return undef;
  }
  $sequence_id = $sh->fetch->[0];
  $sh->finish;

  $self->reordersequence($sequence_id,$newrank);
} 

#
# Subroutine to reorder an annotation in a reading list
#
sub reordernote {
  my $self = shift;
  my($note_id,$newrank) = @_;
  my $sequence_id;

  $note_id = $self->{db}->quote($note_id);
  my $sql = "select sequence_id from sequence where note_id = $note_id";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows < 1) {
      return undef;
  }
  $sequence_id = $sh->fetch->[0];
  $sh->finish;

  $self->reordersequence($sequence_id,$newrank);
} 

#
# Subroutine to reorder a link in a reading list
#
sub reorderlink {
  my $self = shift;
  my($link_id,$newrank) = @_;
  my $sequence_id;

  $link_id = $self->{db}->quote($link_id);
  my $sql = "select sequence_id from sequence where link_id = $link_id";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows < 1) {
      return undef;
  }
  $sequence_id = $sh->fetch->[0];
  $sh->finish;

  $self->reordersequence($sequence_id,$newrank);
} 

#
# Subroutine to do a bulk reordering of a list
#
sub bulkrerank {
    my $self = shift;
    my($modulecode,$newrank,$rerank_type,@picked) = @_;

    # Get the list_id of the reading list that this sequence is in 
    $modulecode = $self->{db}->quote($modulecode);
    my $sql = "select list_id from readinglist where " .
	      "module_code = $modulecode";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    return undef if($sh->rows < 1);
    my @row = $sh->fetchrow_array;
    my $list_id = $row[0];
    $sh->finish;

    # Get <old_ranking,sequence_id>s of picked items in ranked order
    my $conj = "";
    my @pickedrank;
    my @pickedseqid;
    my %pickedsequenceid;
    if($rerank_type eq "alpha") {
        # To sort items alphabetically during reranking we need to get
        # get the annotations out first...
	$sql = "select sequence_id, rank from sequence where " .
	    "list_id = $list_id and (";
        $conj = "";
	foreach my $thing (@picked) {
	    my($thing_type,$thing_id) = split(":",$thing);
	    if($thing_type eq "n") {
		$sql .= " $conj note_id = $thing_id ";
                $conj = "or";
	    }
	}
        $sql .= ")";
        if($conj ne "") {
	    warn "sql = $sql\n";
	    $sh = $self->{db}->prepare($sql);
	    $sh->execute;
	    return undef if($sh->rows < 1);
	    while(@row = $sh->fetchrow_array) {
		my $thisseqid = $self->{db}->quote($row[0]);
		my $thisrank = $row[1];
		push @pickedrank, $thisrank;
		push @pickedseqid, $thisseqid;
		$pickedsequenceid{$thisseqid} = $thisrank;
	    }
	    $sh->finish;
	}

        # ... then the links ...
	$sql = "select sequence_id, rank from sequence where " .
	    "list_id = $list_id and (";
        $conj = "";
	foreach my $thing (@picked) {
	    my($thing_type,$thing_id) = split(":",$thing);
	    if($thing_type eq "l") {
		$sql .= " $conj link_id = $thing_id ";
                $conj = "or";
	    }
	}
        $sql .= ")";
        if($conj ne "") {
	    warn "sql = $sql\n";
	    $sh = $self->{db}->prepare($sql);
	    $sh->execute;
	    return undef if($sh->rows < 1);
	    while(@row = $sh->fetchrow_array) {
		my $thisseqid = $self->{db}->quote($row[0]);
		my $thisrank = $row[1];
		push @pickedrank, $thisrank;
		push @pickedseqid, $thisseqid;
		$pickedsequenceid{$thisseqid} = $thisrank;
	    }
	    $sh->finish;
	}

        # ... then the picked materials out by author name.
	$sql = "select sequence.sequence_id, sequence.rank " .
               "from sequence, material where " .
	       "sequence.list_id = $list_id and ".
               "sequence.material_id = material.material_id " .
               "and ( ";
        $conj = "";
	foreach my $thing (@picked) {
	    my($thing_type,$thing_id) = split(":",$thing);
	    if($thing_type eq "m") {
		$sql .= " $conj sequence.material_id = $thing_id ";
                $conj = "or";
	    }
	}
        $sql .= ") order by concat(material.partauthor, material.authors, ";
        $sql .= "material.parttitle, material.title)";
        if($conj ne "") {
	    warn "sql = $sql\n";
	    $sh = $self->{db}->prepare($sql);
	    $sh->execute;
	    while(@row = $sh->fetchrow_array) {
		my $thisseqid = $self->{db}->quote($row[0]);
		my $thisrank = $row[1];
		push @pickedrank, $thisrank;
		push @pickedseqid, $thisseqid;
		$pickedsequenceid{$thisseqid} = $thisrank;
	    }
	    $sh->finish;
	}

    } else {
	$sql = "select sequence_id, rank from sequence where " .
	    "list_id = $list_id and (";
	foreach my $thing (@picked) {
	    my($thing_type,$thing_id) = split(":",$thing);
	    if($thing_type eq "m") {
		$sql .= " $conj material_id = $thing_id ";
	    } elsif ($thing_type eq "n") {
		$sql .= " $conj note_id = $thing_id ";
	    } else  {
		$sql .= " $conj link_id = $thing_id ";
	    }
	    $conj = "or";
	}
        $sql .= ") order by rank";
        warn "sql = $sql\n";
        $sh = $self->{db}->prepare($sql);
        $sh->execute;
        return undef if($sh->rows < 1);
        while(@row = $sh->fetchrow_array) {
	    my $thisseqid = $self->{db}->quote($row[0]);
	    my $thisrank = $row[1];
	    push @pickedrank, $thisrank;
	    push @pickedseqid, $thisseqid;
	    $pickedsequenceid{$thisseqid} = $thisrank;
	}
	$sh->finish;
    }

    # Get list of <old_ranking,sequence_id> pairs from the database.
    $sql = "select sequence_id, rank from sequence " . 
	" where list_id = $list_id order by rank";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    return undef if($sh->rows < 1);
    my @oldrank;
    my @oldseqid;
    my %allsequenceid;
    while(@row = $sh->fetchrow_array) {
	my $thisseqid = $self->{db}->quote($row[0]);
	my $thisrank = $row[1];
        push @oldrank, $thisrank;
        push @oldseqid, $thisseqid;
        $allsequenceid{$thisseqid} = $thisrank;
    }
    $sh->finish;

    # Our current ranking cursor (so that we normalise everything)
    my $rankcursor = 1;

    # Look at our list from above and normalise ranking until we hit
    # the insertion point.  Ignore items in the picked list
    my $i = 0;
    foreach my $thisrank (@oldrank) {
	last if($thisrank >= $newrank);
        my $thisseqid = $oldseqid[$i++];
        next if($pickedsequenceid{$thisseqid});
        $sql = "update sequence set rank=$rankcursor where " .
	       "sequence_id = $thisseqid";
#        print STDOUT "<p><em>Normalising rank $thisrank to $rankcursor using $sql</em>\n";
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	$sh->finish;
	$rankcursor++;
    }

    # For each picked item, change the rank stored in the data to the new
    # current normalised rank cursor value
    foreach my $thisseqid (@pickedseqid) {
        $sql = "update sequence set rank=$rankcursor where " .
	       "sequence_id = $thisseqid";
#        print STDOUT "<p><em>Moving picked $thisseqid to rank $rankcursor using $sql</em>\n";
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	$sh->finish;
	$rankcursor++;
    }

    # Continue along list from above and normalise ranking from
    # the insertion point onwards  Ignore items in the picked list
    $i = 0;
    foreach my $thisrank (@oldrank) {
        my $thisseqid = $oldseqid[$i++];
	next if($thisrank < $newrank);
#	print STDOUT "<p>thisrank = $thisrank, newrank = $newrank</p>";
        next if($pickedsequenceid{$thisseqid});
        $sql = "update sequence set rank=$rankcursor where " .
	       "sequence_id = $thisseqid";
#        print STDOUT "<p><em>Finishing with normalising rank $thisrank to $rankcursor using $sql</em>\n";
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	$sh->finish;
	$rankcursor++;
    }
}


#
# Subroutine to reorder a sequence enty in a reading list
#
sub reordersequence {
  my $self = shift;
  my($sequence_id,$newrank) = @_;

# warn "in reordersequence($sequence_id,$newrank)\n";

  # Get the list_id of the reading list that this sequence is in 
  my $sql = "select list_id, rank from sequence where " .
    "sequence_id = $sequence_id";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  return undef if($sh->rows < 1);
  my @row = $sh->fetchrow_array;
  my $list_id = $row[0];
  my $oldrank = $row[1];
  $sh->finish;

  # No timewasters please... :-)
  return(1) if($oldrank == $newrank);

  # Find out what the current maximum ranking is
  $list_id = $self->{db}->quote($list_id);
  $sql = "select max(rank) from sequence where list_id = $list_id";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  return undef if($sh->rows < 1);
  my $maxrank = $sh->fetch->[0];
  $sh->finish;

  # Check that the new ranking is sane (and make it so if it isn't)
  $newrank = int($newrank);
  $newrank = 1 if($newrank < 1);
  $newrank = $maxrank if($newrank > $maxrank);

  # Reorder the reading list
  $sequence_id = $self->{db}->quote($sequence_id);
  $sql = "select sequence_id,rank from sequence where list_id = $list_id " .
    "order by rank";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows < 1) {
#    warn "nothing to resequence!\n";
    return undef;
  }
  my $rankcursor = 0;
  my $realnewrank = -1;
  my $tsh;
  while(@row = $sh->fetchrow_array) {
    my $thisseqid = $self->{db}->quote($row[0]);
    my $thisrank = $row[1];
    $rankcursor++;
    # ignore our sequence_id
    next if($thisseqid eq $sequence_id);
    # ignore sequence_id's that are outside of the range that could be
    # reordered by our sequence_id's reranking
    if(($thisrank < $newrank) && ($thisrank < $oldrank)) {
#      warn "$thisrank is less than both $newrank and $oldrank\n";
	if($thisrank != $rankcursor) {
	    $sql = "update sequence set rank=$rankcursor where " .
		"sequence_id = $thisseqid";
	    $tsh = $self->{db}->prepare($sql);
	    $tsh->execute;
	    $tsh->finish;
	}
	next;
    }
    if(($thisrank > $newrank) && ($thisrank > $oldrank)) {
#      warn "$thisrank is less than both $newrank and $oldrank\n";
	if($thisrank != $rankcursor) {
	    $sql = "update sequence set rank=$rankcursor where " .
		"sequence_id = $thisseqid";
	    $tsh = $self->{db}->prepare($sql);
	    $tsh->execute;
	    $tsh->finish;
	}
	next;
    }
    # Do we move this sequence_id up or down in the ranking?
    if($newrank < $oldrank) {
#      warn "$newrank less than $oldrank so inc $thisrank\n";
      $thisrank = $rankcursor + 1;
      $realnewrank = $rankcursor if($realnewrank == -1);
    } elsif($newrank > $oldrank) {
#      warn "$newrank greater than $oldrank so dec $thisrank\n";
      $thisrank = $rankcursor - 1 ;
      $realnewrank = $rankcursor;
    }
    # Do the reordering for this sequence_id
    $thisrank = $self->{db}->quote($thisrank);
    $sql = "update sequence set rank=$thisrank where " .
      "sequence_id = $thisseqid"; 
    $tsh = $self->{db}->prepare($sql);
    $tsh->execute;
    $tsh->finish;
 }
  $sh->finish;
  $realnewrank = $newrank if($realnewrank == -1);

  # OK, once we've jiggled everything else about, put our sequence_id in
  # the right place in the ranking.      
  $realnewrank = $self->{db}->quote($realnewrank);
  $sql = "update sequence set rank=$realnewrank where " .
      "sequence_id = $sequence_id";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  $sh->finish;

# warn "Leaving reordersequence\n";
} 

#
# Subroutine to delete a note from a reading list
#
sub deletenote {
  my $self = shift;
  my($note_id) = @_;

  $note_id = $self->{db}->quote($note_id);
  my $sql = "select list_id from sequence where " .
    "note_id = $note_id";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  my $list_id = undef;
  if($sh->rows > 0) {
    my @row = $sh->fetchrow_array;
    $list_id = $row[0];
  }
  $sh->finish;

  # Delete the actual annotation
  $sql = "delete from annotation where note_id = $note_id";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  $sh->finish;

  # Delete the sequence table entry for this annotation
  $sql = "delete from sequence where note_id = $note_id";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  $sh->finish;
  
  # Re-rank the remaining links, annotations and items in the reading list
  $self->normaliseranking($list_id);    

  return("");
}

#
# Subroutine to delete a link from a reading list
#
sub deletelink {
  my $self = shift;
  my($link_id) = @_;

  $link_id = $self->{db}->quote($link_id);
  my $sql = "select list_id from sequence where " .
    "link_id = $link_id";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  my $list_id = undef;
  if($sh->rows > 0) {
    my @row = $sh->fetchrow_array;
    $list_id = $row[0];
  }
  $sh->finish;

  # Delete the actual link
  $sql = "delete from link where link_id = $link_id";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  $sh->finish;

  # Delete the sequence table entry for this link
  $sql = "delete from sequence where link_id = $link_id";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  $sh->finish;
  
  # Re-rank the remaining links, annotations and items in the reading list
  $self->normaliseranking($list_id);    

  return("");
}

#
# Subroutine to delete an item from a reading list
#
sub deleteitem {
  my $self = shift;
  my($material_id) = @_;

  # The first thing to do is to make a copy of the existing record entry in 
  # the archive_material table so that librarians can get a historical view
  # of what changes have been made to an item.
  $material_id = $self->{db}->quote($material_id);
  my $sql = "insert into archive_material " .
    "(material_id, title, authors, edition, publisher, pubdate, " .
    "control_no, parttitle, partauthor, volume, issue, pages, " .
    "indicative, keytext, purchase, held, openurl, rec_long_loan, " .
    "rec_week_loan, rec_short_loan, rec_reference, url, created_time, " .
    "created_by, modified_time, modified_by) " .
    "select material_id, title, authors, edition, publisher, pubdate, " .
    "control_no, parttitle, partauthor, volume, issue, pages, " .
    "indicative, keytext, purchase, held, openurl, rec_long_loan, " .
    "rec_week_loan, rec_short_loan, rec_reference, url, created_time, " .
    "created_by, modified_time, modified_by " .
    " from material where " .
    "(material_id = " . $material_id . ")";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  $sh->finish;

  # Delete the actual material data
  $sql = "delete from material where material_id = $material_id";
#  warn "<em>sql = $sql</em>\n";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  $sh->finish;

  # Get the note_id and list_id from the sequence table that is related 
  # to this material
  $sql = "select note_id, list_id from sequence where " .
    "material_id = $material_id";
#  warn "<em>sql = $sql</em>\n";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  my $note_id = undef;
  my $list_id = undef;
  if($sh->rows > 0) {
    my @row = $sh->fetchrow_array;
    $note_id = $row[0];
    $list_id = $row[1];
  }
  $sh->finish;

  # If there is a note, delete it.
  if(defined($note_id)) {
    $note_id = $self->{db}->quote($note_id);
    $sql = "delete from annotation where note_id = $note_id";
#    warn "<em>sql = $sql</em>\n";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;
  }

  # Delete the sequence table entry for this item
  $sql = "delete from sequence where material_id = $material_id";
#  warn "<em>sql = $sql</em>\n";
  $sh = $self->{db}->prepare($sql);
  $sh->execute;
  $sh->finish;
  
  # Re-rank the remaining links, annotations and items in the reading list
  $self->normaliseranking($list_id);    

  return("");
}

#
# Subroutine to completely delete a readinglist (all items, notes, archived
# material, everything).
#
sub deletelist {
    my $self = shift;
    my $module_code = shift;   
    my $list_id = -1;

    $module_code = $self->{db}->quote($module_code);
    my $sql = "select list_id from readinglist " .
	"where module_code = $module_code";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;

    if($sh->rows > 0) {
	($list_id) = $sh->fetchrow_array;
	$sh->finish;
    } else {
	$sh->finish;
	return(-1);
    }

    $list_id = $self->{db}->quote($list_id);
    $sql = "select sequence_id, material_id, note_id from sequence where " .
	"list_id = $list_id";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows > 0) {
	while(my @row = $sh->fetchrow_array) {
            my $sequence_id = $row[0];
	    my $material_id = $row[1];
	    my $note_id = $row[2];
            my $newsql = "delete from sequence " .
		"where sequence_id = $sequence_id and " .
		"list_id = $list_id";
	    my $newsh = $self->{db}->prepare($newsql);
	    $newsh->execute;
	    $newsh->finish;
	    if($material_id != 0) {
		$newsql = "delete from archive_material " .
		    "where material_id = $material_id";
		$newsh = $self->{db}->prepare($newsql);
		$newsh->execute;
		$newsh->finish;
		$newsql = "delete from material " .
		    "where material_id = $material_id";
		$newsh = $self->{db}->prepare($newsql);
		$newsh->execute;
		$newsh->finish;
	    }
	    if($note_id != 0) {
		$newsql = "delete from annotation " .
		    "where note_id = $note_id";
		$newsh = $self->{db}->prepare($newsql);
		$newsh->execute;
		$newsh->finish;
	    }
	}
    }
    $sh->finish;
    $sql = "delete from readinglist where " .
	"list_id = $list_id";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;
}

#
# 
#
sub lastupdatedate {
    my $self = shift;
    my ($list_id) = @_;

    my $entry_time = "";
    my $list_time = "";
    $list_id = $self->{db}->quote($list_id);
    my $sql = "select max(modified_time) from sequence where " .
	"list_id = $list_id";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows > 0) {
        ($entry_time) = $sh->fetchrow;
    }
    $sh->execute;
    $sql = "select modified_time from readinglist where " .
	"list_id = $list_id and modified_time >= last_item_update";
    $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows > 0) {
        ($list_time) = $sh->fetchrow;
	$sh->finish;
    } else {
	$sh->finish;
	$sql = "select last_item_update from readinglist where " .
	    "list_id = $list_id";
	$sh = $self->{db}->prepare($sql);
	$sh->execute;
	if($sh->rows > 0) {
	    ($list_time) = $sh->fetchrow;
	}
	$sh->finish;
    }
    my($yearpart,$timepart) = split(/\s+/,$entry_time);
    my($year,$month,$day) = split(/\-/,$yearpart);
    ($yearpart,$timepart) = split(/\s+/,$list_time);
    my($listyear,$listmonth,$listday) = split(/\-/,$yearpart);
    if(($listyear*10000 + $listmonth*100 + $listday) >
       ($year*10000 + $month*100 + $day)) {
        $year = $listyear;
        $month = $listmonth;
        $day = $listday;
    }
    return("$year-$month-$day");
}

#
# Subroutine to normalise the ranking in a reading list
#
sub normaliseranking {
  my $self = shift;
  my ($list_id) = @_;

#  warn "Enterting normaliseranking($list_id)...\n";
  my $newrank = 1;
  $list_id = $self->{db}->quote($list_id);
  my $sql = "select sequence_id, rank from sequence where " .
    "list_id = $list_id order by rank";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows > 0) {
    while(my @row = $sh->fetchrow_array) {
      my $sequence_id = $row[0];
      $sql = "update sequence set rank=$newrank where " .
        "sequence_id = $sequence_id";
#      warn "sql = $sql\n";
      my $tsh = $self->{db}->prepare($sql);
      $tsh->execute;
      $tsh->finish;
      $newrank++;
    }
  }
  $sh->finish;
#  warn "Leaving normaliseranking...\n";
}

sub islibraryassistant {
  my $self = shift;
  my $user = shift;

  $user = $self->{db}->quote($user);
  my $sql = "select list_id from owner where " .
	  "userid = $user and type = 2";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows < 1) {
      return(0);
  } else {
      return(1);
  }
  $sh->finish;
}

sub islibrarian {
  my $self = shift;
  my $user = shift;

  $user = $self->{db}->quote($user);
  my $sql = "select list_id from owner where " .
	  "userid = $user and type = 3";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows < 1) {
      return(0);
  } else {
      return(1);
  }
  $sh->finish;
}

sub librarians {
  my $self = shift;

  my %librarians;
  my $sql = "select owner.userid, owner.email, owner.name, owner.list_id, " .
    "readinglist.module_code, readinglist.module_name " .
    "from owner, readinglist " .
    "where owner.type = 3 and owner.list_id = readinglist.list_id ";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows > 0) {
      # Create the data structure to return.  The structure is:
      #
      # $librarians{userid} = Hash keyed on librarian's user ids
      # $librarians{userid}{useremail} = "Librarian's Email Address";
      # $librarians{userid}{username} = "Librarian's Name";
      # $librarians{userid}{lists}{module_code} 
      #   = Hash of lists keyed on module_codes
      # $librarians{userid}{lists}{module_code}{list_id}
      #   = Module list_id associated with a module code
      # $librarians{userid}{lists}{module_code}{module_name}
      #   = Module name associated with a module_code
      my @row1;
      while(@row1 = $sh->fetchrow_array) {
	  my $userid = $row1[0];
	  my $useremail = $row1[1];
          my $username = $row1[2];
	  my $listid = $row1[3];
          my $module_code = $row1[4];
	  my $module_name = $row1[5];
          $librarians{$userid}->{email} = $useremail;
          $librarians{$userid}->{name} = $username;
          $librarians{$userid}->{lists}->{$module_code}->{list_id} 
            = $listid;
          $librarians{$userid}->{lists}->{$module_code}->{module_name} 
	    = $module_name;

      }
  }
  $sh->finish;
  return(%librarians);
}

sub libraryassistants {
  my $self = shift;

  my %librarians;
  my $sql = "select owner.userid, owner.email, owner.name, owner.list_id, " .
    "readinglist.module_code, readinglist.module_name " .
    "from owner, readinglist " .
    "where owner.type = 2 and owner.list_id = readinglist.list_id ";
  my $sh = $self->{db}->prepare($sql);
  $sh->execute;
  if($sh->rows > 0) {
      # Create the data structure to return.  The structure is:
      #
      # $librarians{userid} = Hash keyed on librarian's user ids
      # $librarians{userid}{useremail} = "Librarian's Email Address";
      # $librarians{userid}{username} = "Librarian's Name";
      # $librarians{userid}{lists}{module_code} 
      #   = Hash of lists keyed on module_codes
      # $librarians{userid}{lists}{module_code}{list_id}
      #   = Module list_id associated with a module code
      # $librarians{userid}{lists}{module_code}{module_name}
      #   = Module name associated with a module_code
      my @row1;
      while(@row1 = $sh->fetchrow_array) {
	  my $userid = $row1[0];
	  my $useremail = $row1[1];
          my $username = $row1[2];
	  my $listid = $row1[3];
          my $module_code = $row1[4];
	  my $module_name = $row1[5];
          $librarians{$userid}->{email} = $useremail;
          $librarians{$userid}->{name} = $username;
          $librarians{$userid}->{lists}->{$module_code}->{list_id} 
            = $listid;
          $librarians{$userid}->{lists}->{$module_code}->{module_name} 
	    = $module_name;

      }
  }
  $sh->finish;
  return(%librarians);
}

#
# Subroutine to allow owners to be searched on by name
#
sub ownernamesearch {
    my $self = shift;
    my($ownername, $all) = @_;
    $ownername = $self->{db}->quote("%$ownername%");
    my $sql = "select owner.userid,owner.list_id,owner.name, ".
      "readinglist.module_code, readinglist.module_name, ". 
      "readinglist.year, readinglist.dept_code, " .
      "readinglist.stage, readinglist.code, readinglist.publish " .
      "from owner, readinglist ".
      "where owner.name like $ownername and owner.type = 0 ".
	  "and readinglist.list_id = owner.list_id order by ".
          "substring_index(substring_index(owner.name, ' ', -1), '.', -1)," .
          "if((ascii(right(owner.userid,1)) < 58) , ".
          "  substring(owner.userid,3,length(owner.userid)-4), ".
          "  substring(owner.userid,3,length(owner.userid)-3)), ".
          "owner.userid,".
          "readinglist.module_code";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my @result;
    my $count = 0;
    my $year = (gmtime(time - 14256000))[5];
#    if($year > 99) {
#	$year = $year - 100;
#	$year = "0".$year if(length($year) == 1);
#    }
    $year+=1900;

    while(my @row =$sh->fetchrow_array) {
        next if($all ne "YES" && $row[5] !~ /^$year/);
	$result[$count]->{userid} = $row[0];
	$result[$count]->{listid} = $row[1];
	$result[$count]->{ownername} = $row[2];
        $result[$count]->{modulecode} = $row[3];
        $result[$count]->{modulename} = $row[4];
        $result[$count]->{year} = $row[5];
        $result[$count]->{dept_code} = $row[6];
        $result[$count]->{stage} = $row[7];
        $result[$count]->{code} = $row[8];
        $result[$count]->{publish} = $row[9];
	$count++;
    }
    $sh->finish;
    return(@result);
}

#
# Subroutine to allow modules to be searched on by name
#
sub modulenamesearch {
    my $self = shift;
    my($modulename, $all) = @_;
    $modulename = $self->{db}->quote("%$modulename%");
    my $sql = "select owner.userid,owner.list_id,owner.name, ".
      "readinglist.module_code, readinglist.module_name, ". 
      "readinglist.year, readinglist.dept_code, " .
      "readinglist.stage, readinglist.code, readinglist.publish " .
      "from owner, readinglist ".
      "where readinglist.module_name like $modulename and owner.type = 0 ".
	  "and readinglist.list_id = owner.list_id order by ".
          "substring_index(substring_index(owner.name, ' ', -1), '.', -1)," .
          "if((ascii(right(owner.userid,1)) < 58) , ".
          "  substring(owner.userid,3,length(owner.userid)-4), ".
          "  substring(owner.userid,3,length(owner.userid)-3)), ".
          "owner.userid,".
          "readinglist.module_code";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my @result;
    my $count = 0;
    my $year = (gmtime(time - 14256000))[5];
#    if($year > 99) {
#	$year = $year - 100;
#	$year = "0".$year if(length($year) == 1);
#    }
    $year+=1900;

    while(my @row =$sh->fetchrow_array) {
        next if($all ne "YES" && $row[5] !~ /^$year/);
	$result[$count]->{userid} = $row[0];
	$result[$count]->{listid} = $row[1];
	$result[$count]->{ownername} = $row[2];
        $result[$count]->{modulecode} = $row[3];
        $result[$count]->{modulename} = $row[4];
        $result[$count]->{year} = $row[5];
        $result[$count]->{dept_code} = $row[6];
        $result[$count]->{stage} = $row[7];
        $result[$count]->{code} = $row[8];
        $result[$count]->{publish} = $row[9];
	$count++;
    }
    $sh->finish;
    return(@result);
}

#
# Subroutine to allow modules to be searched on by code
#
sub modulecodesearch {
    my $self = shift;
    my($modulecode, $all) = @_;
    $modulecode = $self->{db}->quote("%$modulecode%");
    my $sql = "select owner.userid,owner.list_id,owner.name, ".
      "readinglist.module_code, readinglist.module_name, ". 
      "readinglist.year, readinglist.dept_code, " .
      "readinglist.stage, readinglist.code, readinglist.publish " .
      "from owner, readinglist ".
      "where readinglist.module_code like $modulecode and owner.type = 0 ".
	  "and readinglist.list_id = owner.list_id order by " .
          "substring_index(substring_index(owner.name, ' ', -1), '.', -1)," .
          "if((ascii(right(owner.userid,1)) < 58) , ".
          "  substring(owner.userid,3,length(owner.userid)-4), ".
          "  substring(owner.userid,3,length(owner.userid)-3)), ".
          "owner.userid,".
          "readinglist.module_code";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    my @result;
    my $count = 0;
    my $year = (gmtime(time - 14256000))[5];
#    if($year > 99) {
#        $year = $year - 100;
#	$year = "0".$year if(length($year) == 1);
#    }
    $year+=1900;

    while(my @row =$sh->fetchrow_array) {
        next if($all ne "YES" && $row[5] !~ /^$year/);
	$result[$count]->{userid} = $row[0];
	$result[$count]->{listid} = $row[1];
	$result[$count]->{ownername} = $row[2];
        $result[$count]->{modulecode} = $row[3];
        $result[$count]->{modulename} = $row[4];
        $result[$count]->{year} = $row[5];
        $result[$count]->{dept_code} = $row[6];
        $result[$count]->{stage} = $row[7];
        $result[$count]->{code} = $row[8];
        $result[$count]->{publish} = $row[9];
	$count++;
    }
    $sh->finish;
    return(@result);
}

#
# Subroutine to tell us if a list has changed since a given date (and return
# when the last change was and by who if it has)
#
sub changed {
    my $self = shift;
    my($list_id,$startdate) = @_;

    $startdate = $self->{db}->quote($startdate);
    my $bywho = "";
    my $sql = "select last_item_update, last_item_update_by from readinglist ".
    "where last_item_update > $startdate and list_id=$list_id";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows > 0) {
	($startdate,$bywho) = $sh->fetchrow_array;
    } else {
	$startdate = "";
    }
    $sh->finish;
    return($startdate,$bywho);
}

#
# Subroutine to give us a number of list/number of items for a given department
# in a given year
#
sub deptbreakdown {
    my $self = shift;
    my($deptcode,$year) = @_;

    my $lists = 0;
    my $items = 0;    
    my @ret = ();

    my $qyear = $self->{db}->quote("$year");
    my $qdeptcode = $self->{db}->quote("$deptcode");

    my $sql = "select count(list_id) from readinglist ".
	"where year=$qyear and dept_code=$qdeptcode";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows > 0) {
	($lists) = $sh->fetchrow_array;
    }
    $sh->finish;
    if($lists) {
	$sql = "select count(sequence.material_id) " .
           "from sequence, readinglist " .
           "where sequence.material_id != \"\" " .
           "and sequence.list_id = readinglist.list_id " .
           "and readinglist.year = $qyear" .
           "and readinglist.dept_code = $qdeptcode";
        $sh = $self->{db}->prepare($sql);
	$sh->execute;
	if($sh->rows > 0) {
	    ($items) = $sh->fetchrow_array;
	}
        $sh->finish;
    }
    @ret = ($lists,$items);
    return(@ret);
}

#
# Subroutine to delete a user as an owner from a reading list
#
sub deleteowner {
    my $self = shift;
    my($list_id,$userid) = @_;

    $list_id = $self->{db}->quote($list_id);
    $userid = $self->{db}->quote($userid);

    my $sql = "delete from owner ".
        "where list_id=$list_id and userid=$userid";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;
}

#
# Subroutine to timestamp a reading list modified_time/modified_by field
#
sub modify_timestamp {
    my $self = shift;
    my($module_code) = @_;

    my $qmodule_code = $self->{db}->quote($module_code);
    my $sql = "update readinglist set modified_time=now() where ".
        "module_code = $qmodule_code";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;
}

#
# Subroutine to turn on the publishing flag for a reading list
#
sub publish {
    my $self = shift;
    my($module_code) = @_;

    my $qmodule_code = $self->{db}->quote($module_code);
    my $sql = "update readinglist set publish='Y' where ".
        "module_code = $qmodule_code";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;
}

#
# Method to mark a reading list as unpublished
#
sub unpublish {
    my $self = shift;
    my($module_code) = @_;

    my $qmodule_code = $self->{db}->quote($module_code);
    my $sql = "update readinglist set publish='N' where ".
        "module_code = $qmodule_code";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    $sh->finish;
}

#
# Method to return a list of available stages
#
sub liststages {
    my $self = shift;
    my @stages = ();

    my $sql = "select id,code,name,description from stage";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;

    my $count = 0;
    while(my @row = $sh->fetchrow_array) {
	my($id, $code, $name, $description) = @row;
        $stages[$count]->{id} = $id;
        $stages[$count]->{code} = $code;
        $stages[$count]->{name} = $name;
        $stages[$count]->{description} = $description;
	$count++;
    }
    $sh->finish;
    return(@stages);
}

#
# Method to return a list of available stages for a given year and
# department that can be displayed to end users
#
sub displaystages {
    my $self = shift;
    my($dept,$year) = @_;
    
    my @stages = ();

    my $qyear = $self->{db}->quote($year);
    my $qdept = $self->{db}->quote($dept);
    my $sql = "select distinct(stage.code), stage.name, stage.description " .
      " from stage, readinglist where " .
      " stage.code = readinglist.stage and " .
      " stage.browsable = 'Y' and " .
      " readinglist.dept_code = $qdept and " .
      " readinglist.year = $qyear and " .
      " readinglist.publish = 'Y' " .
      " order by stage.code ";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;

    my $count = 0;
    while(my @row = $sh->fetchrow_array) {
	my($code, $name, $description) = @row;
        $stages[$count]->{code} = $code;
        $stages[$count]->{name} = $name;
        $stages[$count]->{description} = $description;
        $count++;
    }
    $sh->finish;
    return(@stages);
}

#
# Method to return a list of available years for a given department
# that can be displayed to end users
#
sub displayyears {
    my $self = shift;
    my($dept) = @_;
    
    my @years = ();

    my $qdept = $self->{db}->quote($dept);
    my $sql = "select distinct(readinglist.year) " .
      " from readinglist where " .
      " readinglist.dept_code = $qdept and " .
      " readinglist.publish = 'Y' " .
      " order by readinglist.year desc";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;

    my $count = 0;
    while(my @row = $sh->fetchrow_array) {
	$years[$count] = $row[0];
        $count++;
    }
    $sh->finish;
    return(@years);
}

sub extract_module_info {
    my $self = shift;
    my ($modulecode) = shift;

    my $modulecodeformat = ReadingLists::ModuleCodeFormat;
    
    # Extract year portion of the module code
    my $autoyear = "";
    my $autoyear_ok = 0;
    my $count = -1;
    foreach my $char (split(//,$modulecodeformat)) {
	$count++;
        next if($char ne "Y");
        $autoyear .= substr($modulecode,$count,1);
	$autoyear_ok = 1;
    }
    if(length($autoyear) == 2) {
	if($autoyear == 98 || $autoyear == 99) {
	    $autoyear += 1900;
	} else {
	    $autoyear += 2000;
	}
    }

    # Extract the department code portion of the module code
    my $autodept_code = "";
    my $autodept_code_ok = 0;
    $count = -1;
    foreach my $char (split(//,$modulecodeformat)) {
	$count++;
        next if($char ne "D");
        $autodept_code .= substr($modulecode,$count,1);
	$autodept_code_ok = 1;
    }

    # Extract the stage portion of the module code
    my $autostage = "";
    my $autostage_ok = 0;
    $count = -1;
    foreach my $char (split(//,$modulecodeformat)) {
	$count++;
        next if($char ne "S");
        $autostage .= substr($modulecode,$count,1);
	$autostage_ok = 1;
    }

    # Extract the opaque code portion of the module code
    my $autocode = "";
    my $autocode_ok = 0;
    $count = -1;
    foreach my $char (split(//,$modulecodeformat)) {
	$count++;
        next if($char ne "X");
        $autocode .= substr($modulecode,$count,1);
	$autocode_ok = 1;
    }


    return($autoyear,$autodept_code,$autostage,$autocode,
	   $autoyear_ok,$autodept_code_ok,$autostage_ok,$autocode_ok);
}

#
# Method to find the next free module code.  This only works if there
# aren't gaps in your module code sequence and of course doesn't know
# anything about module codes that your central admin have allocated
# to people but which haven't been used yet.
#
sub find_next_module_code {
    my $self = shift;
    my $year = shift;
    my $dept_code = shift;
    my $stage = shift;

    my $newopaquecode = undef;

    my $sql = "select max(code)+1 from readinglist where year='$year' " .
	"and dept_code='$dept_code' and stage='$stage'";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows > 0){
        # Get the new largest opaque code and check to make sure that its
        # length in characters is <= the length allowed for the opaque code
        # in the module code format.
	$newopaquecode = ($sh->fetchrow_array)[0];
        my $format = $self->ModuleCodeFormat;
        $format =~ s/[^X]//g;
	if(length($newopaquecode) > length($format)) {
	    $newopaquecode = undef;
	} elsif (length($newopaquecode) < length($format)) {
	    $newopaquecode = "0"x(length($format) - length($newopaquecode)) .
		$newopaquecode;
	}
    }
    $sh->finish;

    return($newopaquecode);
}

#
# Method to get the user's name, type and email address based on the last
# set of values that they gave
#
sub prior_user_details {
    my $self = shift;
    my $me = shift;

    my $results;
    @{$results} = ('No Name', 0, 'library@lboro.ac.uk');
    my $qme = $self->{db}->quote($me);
    my $sql = "select name, type, email from owner where userid=$qme order by modified_time desc limit 1";
    warn "SQL = $sql\n";
    my $sh = $self->{db}->prepare($sql);
    $sh->execute;
    if($sh->rows > 0) {
	$results = $sh->fetchrow_arrayref;
	warn "$results->[0]\n";
    }
    $sh->finish;
    return $results;
}

1;

__END__

=head1 NAME

ReadingLists - Module implementing a reading list object and methods

=head1 REVISION

$Revision: 1.86 $

=head1 SYNOPSIS

use ReadingLists;

$readinglist = new ReadingLists;

$dsn = ReadingLists::dsn;

print ReadingLists::bl;

print ReadingLists::br;

print ReadingLists::trailer;

$readinglist->createlist($module_code, $module_name, $userid, $ownername,
  $owneremail, $ownertype, $owneruserid);

$readinglist->updatelist($module_code, $module_name, $userid, $ownername,
  $owneremail, $ownertype,$owneruserid);

$readinglist->copylist($old_module_code, $new_module_code, $new_module_name,
		       $userid);

@departments = $readinglist->departments;

$result = $readinglist->validmodulecode($module_code);

@lists = $readinglist->available_lists;

@owners = $readinglist->list_owners($list_id);

$list_details = $readinglist->list_details($module_code);

@entries = $readinglist->getlistentries($list_id);

$notetext = $readinglist->getnote($note_id);

$note_id = $readinglist->insertnote($module_code,$notetext);

$readinglist->updatenote($note_id, $notetext);

@lists_details = $readinglist->listsowned($userid);

@materials = $readinglist->controlfind($control_number);

$result = $readinglist->reorderitem($material_id, $newrank);

$result = $readinglist->reordernote($note_id, $newrank);

$result = $readinglist->reordersequence($sequence_id, $newrank);

$result = $readinglist->deletenote($note_id);

$result = $readinglist->deleteitem($material_id);

$result = $readinglist->normaliseranking($list_id);

$result = $readinglist->islibrarian($user_id);

$date = $readinglist->changed($list_id,$testdate);

($listcount,$itemcount) = $readinglist->deptbreakdown($deptcode,$year);

$readinglist->deleteowner($listid,$owneruserid);

=head1 DESCRIPTION

This module implements a reading list object and provides a number of methods
for create, destroying and manipulating these objects.  The reading list data
itself is currently stored in a relational database (specifically a MySQL
database) but the object methodology employed by this module will allow
the actual representation of this data to be changed in the future without
affecting scripts that make use of it.

=head1 METHODS

=head2 new

Creates a new reading list object.  It also opens a database connection to
the RDBMS that stores the reading list data, using the dsn method below.

=head2 dsn

Returns a string which consists of the DSN for the backend RDBMS in a format
ready to pass to DBI->connect.

=head2 bl

Returns HTML openning sequence for debugging output - not required for normal
use of the module.

=head2 br

Returns HTML closing sequence for debugging output - not required for normal
use of the module.

=head2 trailer

Returns HTML for closing the HTML page.  This can contain some standard
links to known resources (eg main manus, departmental web pages, OPAC, etc)
that should appear at the end of each CGI script.

=head2 available_lists

Return an array of hashes that contain the details about all of the available 
reading lists that the system knows about.  The keys to the hash are:

=over 4

=item list_id - The unique (hopefully!) ID for this reading list

=item module_code - The code for the module for which this is a reading list

=item module_name - The name of the module for which this is a reading list

=back

=head2 createlist

This method creates the basic data structures in the backend database for
a new reading list and its owner.  It requires the module code, module name,
the userid of the person creating this list, the list owner's real name,
the list owner's email address, what type of owner this is and the owner's
userid.  The owner types are:

=over 4

=item 0 - Normal academic reading list owner

=item 1 - Departmental administrative staff

=item 2 - Library staff

=back

=head2 updatelist

This method takes the same parameters as the createlist method but instead
of creating a new reading list, it updates an existing one.  Note that if
the owner userid that is passed is different to any of the existing
reading list owners, a new owner of the specified type will be added to the
reading list.

=head2 copylist

This method takes the module code of an existing readinglist, the module
code of a new reading list, a name for the new module and a userid and
creates the new reading list (marking it as created by the specified user
id) and then copies the contents of the old reading list over to the new
reading list.

=head2 departments

This method returns a data structure containing information about departments
that may have reading lists.  The data structure is an array of hashes, with
hash keys of:

=over 4

=item dept_id - Unique ID field for this department

=item dept_code - Two letter code for the department

=item dept_name - Name of the department

=back

=head2 validmodulecode

This method takes a module code string as a parameter and checks it for
validity.  It returns 1 if it is valid and 0 if it is not.  Note that validity
checking is location dependent and this routine uses Loughborough University's
module code format by default.

=head2 list_owners

This method takes a parameter which is the ID of a reading_list and
returns an array of hashes detailing the owners of that reading list.
The owners of a reading list can include the academic(s) that actually
owns the list, administrative staff in departments with permission to
update the reading lists and library staff who maintain the book stock
for that reading list.  The keys in the returned hash are:

=over 4

=item name - The user real name

=item userid - The userid of this user

=item type - The type of owner.  0 means a real academic, 1 means departmental
admin staff and 2 means library staff

=item email - The email address of the user

=back 

=head2 list_details

This method takes a module code as a parameter and returns a hash
containing the details of the reading list related to that module.  The
keys for the hash are:

=over 4

=item list_id - The ID of the reading list for that module code

=item module_name - The name of the module with that module code

=item created_time - When the reading list for that module was created

=item modified_time - When the reading list for that module was last
modified.

=back

=head2 controlfind

This method takes a control number (ISBN, ISSN or local control number)
as a parameter and returns an array of details of all entries in any
reading lists that have that control number.  The returned array is
actually an array of hashes with keys of:

=over 4

=item material_id - The material_id of this item,

=item list_id - The list_id of the reading list containing this item,

=item module_code - The module code of the reading list containing this item,

=item module_name - The module name of the reading list containing this item.

=back

=head2 getlistentries

This method takes a reading list ID has its only parameter and returns
an array of hashes contain information about the entries in the reading
list with that ID.  The keys for the hashes are:

=over 4

=item sequence_id - The ID of the sequence entry for this entry

=item material_id - The ID of the material (if any) for this entry

=item note_id - The ID of the annotation (if any) for this entry

=item rank - The ranking of this entry in the reading list sequence

=item created_time - When this reading list sequence was created

=item created_by - The userid of who created this reading list sequence

=item modified_time - When this reading list sequence was last modified

=item modified_by - The userif of who last modified this reading list
sequence

=back

If there are no material items or annotations in a reading list (ie
its completely empty) this method returns undef.

=head2 getnote

Taking a note ID as its parameter, this method returns the text of the
annotation with that ID.  A similar method exists in the ReadingListsItem
module but that only returns a note related to a material item - this
method will return any annotation including the stand alone one.

=head2 insertnote

This method takes two parameters: a module code and the text of an
annotation.  It attempts to create an annotation and insert it into
the sequence for the reading list for the module.  If it fails it
returns undef.

=head2 updatenote($note_id, $notetext);

This method takes three parameters: a note ID and the
text of an annotation.  It attempts to update the annotation with the
specified note ID with the new text.

=head2 listsowned

This methods takes a userid as its only parameter and retuns an array of
hashes as a result.  These hashes contain the details of all the reading
lists of which this user is an owner (in any capacity).  The keys in the
hashes are:

=over 4

=item list_id - The ID of the reading list

=item module_code - The module code for which this is a reading list

=item module_name - The name of the module with that module code

=item created_time - When the reading list for that module was created

=item modified_time - When the reading list for that module was last
modified.

=back

This method retuns undef if there were no reading lists assocated with
the specified userid.

=head2 reorderitem

This method takes the ID of a material item and a new ranking for this
item.  This is really a convenience method in that it looks up the
sequence_id for this item and then calls the reordersequence method.

=head2 reordernote

This method takes the ID of an annotation and a new ranking for this
annotation.  This is really a convenience method in that it looks up the
sequence_id for this item and then calls the reordersequence method.

=head2 reordersequence

This method takes the ID of a sequence entry and a new ranking for that
sequence.  It updates the other entries in the sequence so that the
specified entry has the given new ranking and the other entries are 
reranked around it.

=head2 deletenote

Taking a note ID as a parameter this method is used to delete an annotation
and remove it from the reading list.  This method calls the normaliseranking
method to ensure that the remain entries in the sequence for this reading
list are reranked to take up the space.

=head2 deleteitem

This method takes the ID of a material ID which is to be deleted and removed
from the reading list.  This method calls the normaliseranking method to
ensure that the remain entries in the sequence for this reading list are
reranked to take up the space.

=head2 normaliseranking

This method requires a list ID as its parameter and attempts to
normalise the ranking of the sequence entries in the reading list.
The normalisation removes any gaps in the ranking so that it the ranks
are monotomically increasing from 1.

=head2 islibrarian

A user id is passed as the parameter for this method which returns a
1 if the user is a librarian or 0 otherwise.

=head2 islibraryassistant

A user id is passed as the parameter for this method which returns a
1 if the user is a library assistant or 0 otherwise.

=head2 changed

A list_id and a date are passed as the parameters and the method then returns
either an empty string if the list with the given list_id was not changed
since the supplied date, or the date that the list was last changed on.

=head2 deptbreakdown

This method takes a department code and a two digit year as parameters.
These are used to form module codes for readinglists owned by that department
in that year which is then used to work out the number of readinglists that
that department has in that year and how many items are in those lists.

=head2 deleteowner

This method requires a listid and the userid of a list owner.  It then
deletes that user from being an owner of the given list.

=head1 BUGS

Some of the methods in this module should reallly be in the
ReadingListsItem module as they efectively refer to an item (eg
reorderitem and deleteitem), although as they affect the whole
sequence in a reading list that module would just end up having to
call this one still.

=head1 COPYRIGHT

This code is free software and is distributed under the Gnu General
Public License (GPL).  A copy of the GPL should have been supplied along
with this software.  If you don't have a copy you can get one from
<URL:http://www.gnu.org>.

This code developed by the Systems Team in the Pilkington Library at
Loughborough University, England.  The code as delivered "as is"; no
warrenty is made for its fitness of purpose or against claims of loss
or damage resulting from its use.

=head1 AUTHORS

Jon Knight <J.P.Knight@lboro.ac.uk> and 
Gary Brewerton <G.P.Brewerton@lboro.ac.uk>

