#
# Schema for Loughborough University Reading Lists
#
# Authors: J.P.Knight@lboro.ac.uk and G.P.Brewerton@lboro.ac.uk
#
# $Id: schema.sql,v 1.18 2005/08/22 16:25:43 jon Exp jon $
#

use mysql;
create database readinglist;
use readinglist;

# The material table holds details of a active item in system.  Note that
# whilst different readinglists may reference the same physical book they
# might wish to refer to different parts of it or have different styles of
# author citation.  This means that we will potentially have lots of materials
# with the same control numbers (ISBNs, etc) but with different details in
# this table.  Bear this in mind when interfacing with the librarians and
# Talis, etc.
create table material (
  material_id INT NOT NULL PRIMARY KEY AUTO_INCREMENT,
  title TEXT,
  authors TEXT,
  edition INT,
  publisher TEXT,
  pubdate SMALLINT,
  control_no VARCHAR(15),
  parttitle TEXT,
  partauthor TEXT,
  volume TEXT,
  issue TEXT,
  pages TEXT,
  indicative BOOL,
  keytext BOOL,
  purchase BOOL,
  held INT,
  rec_long_loan INT,
  rec_week_loan INT,
  rec_short_loan INT,
  rec_reference INT,
  url TEXT,
  openurl enum('Y','N'),
  private_note text,
  created_time DATETIME,
  created_by VARCHAR(12),
  modified_time DATETIME,
  modified_by VARCHAR(12)
);

# This is used to archive changes in material entries.  Everytime an academic
# edits the details of an item, it is recorded in here.  Note that we don't
# archive changes to their notes as we don't really care too much about what
# they say in them.  The edit_number column in this table is a monotomically
# increasing number so that we can archive lots of changes for a particular
# material.
create table archive_material (
  material_id INT NOT NULL,
  title TEXT,
  authors TEXT,
  edition INT,
  publisher TEXT,
  pubdate SMALLINT,
  control_no VARCHAR(15),
  parttitle TEXT,
  partauthor TEXT,
  volume TEXT,
  issue TEXT,
  pages TEXT,
  indicative BOOL,
  keytext BOOL,
  purchase BOOL,
  held INT,
  rec_long_loan INT,
  rec_week_loan INT,
  rec_short_loan INT,
  rec_reference INT,
  url TEXT,
  openurl enum('Y','N'),
  private_note text,
  created_time DATETIME,
  created_by VARCHAR(12),
  modified_time DATETIME,
  modified_by VARCHAR(12),
  edit_number INT
);

# This is the table that associates a reading list's internal list_id with
# the module name and code.
create table readinglist (
  list_id INT NOT NULL PRIMARY KEY AUTO_INCREMENT,
  module_code VARCHAR(10),
  module_name VARCHAR(100),
  dept_code CHAR(2),
  year YEAR,
  stage CHAR,
  code VARCHAR(5),
  publish ENUM('Y','N'),
  created_time DATETIME,
  modified_time DATETIME,
  last_item_update DATETIME,
  last_item_update_by VARCHAR(12)
);

# This table is for annotations.  It does double duty for both standalone
# annotations and also notes included alongside specific items.
create table annotation (
  note_id INT NOT NULL PRIMARY KEY AUTO_INCREMENT,
  content TEXT,
  created_time DATETIME,
  created_by VARCHAR(12),
  modified_time DATETIME,
  modified_by VARCHAR(12)
);

# The sequence table is mainly used to order the materials/notes in a reading
# list for display to the users.
create table sequence (
  sequence_id INT NOT NULL PRIMARY KEY AUTO_INCREMENT,
  list_id INT,
  material_id INT,
  note_id INT,
  link_id INT,
  rank INT,
  created_time DATETIME,
  created_by VARCHAR(12),
  modified_time DATETIME,
  modified_by VARCHAR(12),
  number_of_edits INT,
  KEY `list_id` (`list_id`),
  KEY `material_id` (`material_id`)
);

# The owner table is used to record details about the owner(s) of a reading
# list.  We make a note of the type of owner so that librarians and
# administrators can also manipulate reading lists if required.
create table owner (
  list_id INT,
  name VARCHAR(30),
  userid VARCHAR(12) NOT NULL,
  type CHAR,
  email VARCHAR(50),
  modified_time DATETIME
);

# The departments table is used to list details about departments from which
# the staff who own reading lists come.
create table departments (
  dept_id INT NOT NULL PRIMARY KEY AUTO_INCREMENT,
  dept_code CHAR(2),
  dept_name VARCHAR(100),
  dept_faculty VARCHAR(15),
  browsable ENUM('Y','N')
);

# Table for holding information about inter-list links (sublists, etc).
create table link (
  link_id INT NOT NULL PRIMARY KEY AUTO_INCREMENT,
  source_list_id INT NOT NULL,
  target_list_id INT NOT NULL,
  rollover ENUM('Y','N'),
  prelinktext TEXT,
  postlinktext TEXT,
  created_time DATETIME,
  created_by VARCHAR(12),
  modified_time DATETIME,
  modified_by VARCHAR(12)
);

# Table for holding information about course stages
create table stage (
  id		   INT NOT NULL PRIMARY KEY AUTO_INCREMENT,
  code		   VARCHAR(5),
  name		   VARCHAR(100),
  description	   TEXT,
  browsable	   ENUM('Y','N')
);
